import logging
import time
from typing import List, Optional

import pandas as pd
from ouro._resource import SyncAPIResource

log: logging.Logger = logging.getLogger(__name__)


__all__ = ["Datasets"]


class Datasets(SyncAPIResource):
    def create(
        self,
        name: str,
        visibility: str,
        data: Optional[pd.DataFrame] = None,
        monetization: Optional[str] = None,
        price: Optional[float] = None,
        description: Optional[str] = None,
    ):
        df = data.copy()
        # Get a sql safe table name from the name
        table_name = name.replace(" ", "_").lower()

        # Reset the index if it exists to use as the primary key
        index_name = df.index.name
        if index_name:
            df.reset_index(inplace=True)

        create_table_sql = pd.io.sql.get_schema(
            df,
            name=table_name,
            schema="datasets",
            # TODO: Add support for primary keys
            # keys=index_name
        )

        create_table_sql = create_table_sql.replace(
            "TIMESTAMP", "TIMESTAMP WITH TIME ZONE"
        )
        create_table_sql = create_table_sql.replace(
            "CREATE TABLE", "CREATE TABLE IF NOT EXISTS"
        )

        log.debug(f"Creating a dataset:\n{create_table_sql}")

        body = {
            "name": name,
            "visibility": visibility,
            "monetization": monetization,
            "price": price,
            "description": description,
            "schema": create_table_sql,
        }

        # Filter out None values
        body = {k: v for k, v in body.items() if v is not None}

        request = self.client.post(
            "/datasets/create/from-schema",
            json={"dataset": body},
        )
        request.raise_for_status()
        response = request.json()

        log.info(response)

        if response["error"]:
            raise Exception(response["error"])

        # Good response, but no data to insert right now
        if data is None:
            return response["data"]

        # Good response, we can now insert the data
        created = response["data"]
        table_name = created["metadata"]["table_name"]

        insert_data = self._serialize_dataframe(df)
        # Insert the data into the table
        # TODO: May need to batch insert
        insert = self.database.table(table_name).insert(insert_data).execute()
        if len(insert.data) > 0:
            log.info(f"Inserted {len(insert.data)} rows into {table_name}")

        # Update the dataset with a data preview
        update = self.update(created["id"], preview=insert_data[0:7])

        return response["data"]

    def retrieve(self, id: str):
        """
        Retrieve a Dataset by its id
        """
        request = self.client.get(
            f"/datasets/{id}",
        )
        request.raise_for_status()
        response = request.json()
        if response["error"]:
            raise Exception(response["error"])
        return response["data"]

    def query(self, id: str) -> pd.DataFrame:
        """
        Query a Dataset's data by its id
        """
        request = self.client.get(
            f"/datasets/{id}/data",
        )
        request.raise_for_status()
        response = request.json()
        if response["error"]:
            raise Exception(response["error"])
        return pd.DataFrame(response["data"])

    def load(self, table_name: str):
        """
        Load a Dataset's data by its table name. Good for large datasets.
        Method checks the row count and loads the data in batches if it's too big.
        """
        start = time.time()

        row_count = self.database.table(table_name).select("*", count="exact").execute()
        row_count = row_count.count

        log.info(f"Loading {row_count} rows from datasets.{table_name}...")
        # Batch load the data if it's too big
        if row_count > 1_000_000:
            data = []
            for i in range(0, row_count, 1_000_000):
                log.debug(f"Loading rows {i} to {i+1_000_000}")
                res = (
                    self.database.table(table_name)
                    .select("*")
                    .range(i, i + 1_000_000)
                    .execute()
                )
                data.extend(res.data)
        else:
            res = self.database.table(table_name).select("*").limit(1_000_000).execute()
            data = res.data

        end = time.time()
        log.info(f"Finished loading data in {round(end - start, 2)} seconds.")

        self.data = data
        return data

    def schema(self, id: str):
        """
        Retrieve a Dataset's schema
        """
        request = self.client.get(
            f"/datasets/{id}/schema",
        )
        request.raise_for_status()
        response = request.json()
        if response["error"]:
            raise Exception(response["error"])
        return response["data"]

    def update(
        self,
        id: str,
        name: Optional[str] = None,
        visibility: Optional[str] = None,
        description: Optional[str] = None,
        preview: Optional[List[dict]] = None,
        data: Optional[pd.DataFrame] = None,
        monetization: Optional[str] = None,
        price: Optional[float] = None,
    ):
        """
        Update a Dataset's data by its id
        """

        body = {
            "name": name,
            "visibility": visibility,
            "monetization": monetization,
            "price": price,
            "description": description,
            "preview": preview,
        }
        # Filter out None values
        body = {k: v for k, v in body.items() if v is not None}

        request = self.client.put(
            f"/datasets/{id}",
            json={"dataset": body},
        )
        request.raise_for_status()
        response = request.json()
        if response["error"]:
            raise Exception(response["error"])

        # Make the data update if it's provided
        if data is not None:
            table_name = self.retrieve(id)["metadata"]["table_name"]
            insert_data = self._serialize_dataframe(data)

            insert = self.database.table(table_name).insert(insert_data).execute()
            if len(insert.data) > 0:
                log.info(f"Inserted {len(insert.data)} rows into {table_name}")

        return response["data"]

    def _serialize_dataframe(self, data: pd.DataFrame) -> List[dict]:
        """
        Make a DataFrame serializable by converting NaN values to None,
        formatting datetime columns to strings, and converting empty strings to None.
        """
        clean = data.copy()

        # Fill NaN values with None
        clean = clean.where(pd.notnull(clean), None)
        clean = clean.map(lambda x: None if pd.isna(x) or x == "" else x)

        # Convert datetime columns to strings
        for column in clean.columns:
            if clean[column].dtype == "datetime64[ns]":
                clean[column] = clean[column].dt.strftime("%Y-%m-%d")

        clean = clean.to_dict(orient="records")
        # Ensure that we're not inserting any NaN values by converting them to None
        clean = [
            {k: v if not pd.isna(v) else None for k, v in row.items()} for row in clean
        ]

        return clean



from __future__ import annotations

import logging
import os
from typing import Any, Union

import httpx
import supabase
from ouro.config import Config
from ouro.resources import Conversations, Datasets, Files, Posts
from supabase.client import ClientOptions
from typing_extensions import override

from .__version__ import __version__
from ._constants import (  # DEFAULT_TIMEOUT,; MAX_RETRY_DELAY,; INITIAL_RETRY_DELAY,; RAW_RESPONSE_HEADER,; OVERRIDE_CAST_TO_HEADER,; DEFAULT_CONNECTION_LIMITS,
    DEFAULT_MAX_RETRIES,
)
from ._exceptions import APIStatusError, OuroError
from ._utils import is_mapping  # is_given,

__all__ = ["Ouro"]


log: logging.Logger = logging.getLogger(__name__)


class OuroAuth(httpx.Auth):
    requires_response_body = True

    def __init__(self, api_key, refresh_url):
        self.api_key = api_key
        self.refresh_url = refresh_url
        self.access_token = None

    def auth_flow(self, request):
        # If we don't have an access token, we need to get one
        if not self.access_token:
            refresh_response = yield self.build_refresh_request()
            self.update_tokens(refresh_response)

        request.headers["Authorization"] = self.access_token
        yield request

    def build_refresh_request(self):
        # Return an `httpx.Request` for refreshing tokens.
        return httpx.Request("POST", self.refresh_url, json={"pat": self.api_key})

    def update_tokens(self, response):
        data = response.json()
        self.access_token = data["token"]


class Ouro:
    # Resources
    datasets: Datasets
    files: Files
    posts: Posts
    conversations: Conversations

    # Client options
    api_key: str
    organization: str | None
    project: str | None

    # Clients
    client: httpx.Client
    database: supabase.Client  # datasets schema
    supabase: supabase.Client  # public schema

    # Auth config
    access_token: str | None

    # Connection options
    database_url: str | httpx.URL | None
    database_anon_key: str | None
    base_url: str | httpx.URL | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        organization: str | None = None,
        project: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, None] = None,
        # max_retries: int = DEFAULT_MAX_RETRIES,
        # default_headers: Mapping[str, str] | None = None,
        # default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        # http_client: httpx.Client | None = None,
    ) -> None:
        """Construct a new synchronous ouro client instance.

        This automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `api_key` from `OURO_API_KEY`
        - `organization` from `OURO_ORG_ID`
        - `project` from `OURO_PROJECT_ID`
        """
        if api_key is None:
            api_key = os.environ.get("OURO_API_KEY")
        if api_key is None:
            raise OuroError(
                "The api_key client option must be set either by passing api_key to the client or by setting the OURO_API_KEY environment variable"
            )
        self.api_key = api_key

        if organization is None:
            organization = os.environ.get("OURO_ORG_ID")
        self.organization = organization

        if project is None:
            project = os.environ.get("OURO_PROJECT_ID")
        self.project = project

        # Set config for Supabase client and Ouro client
        self.base_url = Config.OURO_BACKEND_URL
        self.database_url = Config.SUPABASE_URL
        self.database_anon_key = Config.SUPABASE_ANON_KEY

        # Make a private property for the access token
        self.access_token = None
        self.refresh_token = None
        # Create the httpx client
        self.client = httpx.Client(
            auth=OuroAuth(api_key, refresh_url=self.base_url + "/users/get-token"),
            base_url=self.base_url,
            timeout=timeout,
        )

        # Run the login flow
        self.login()

        # Initialize resources
        # self.users = Users(self)
        self.datasets = Datasets(self)
        self.files = Files(self)
        self.posts = Posts(self)
        self.conversations = Conversations(self)

    # @property
    # @override
    # def auth_headers(self) -> dict[str, str]:
    #     api_key = self.api_key
    #     return {"Authorization": f"Bearer {api_key}"}

    # @property
    # @override
    # def default_headers(self) -> dict[str, str | Omit]:
    #     return {
    #         **super().default_headers,
    #         "Ouro-Organization": (
    #             self.organization if self.organization is not None else Omit()
    #         ),
    #         "Ouro-Project": self.project if self.project is not None else Omit(),
    #         **self._custom_headers,
    #     }

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        data = body.get("error", body) if is_mapping(body) else body
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=data)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(
                err_msg, response=response, body=data
            )

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(
                err_msg, response=response, body=data
            )

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=data)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=data)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(
                err_msg, response=response, body=data
            )

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=data)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(
                err_msg, response=response, body=data
            )
        return APIStatusError(err_msg, response=response, body=data)

    def login(self):
        url: str = self.database_url
        key: str = self.database_anon_key

        api_key = self.api_key

        # Send a request to Ouro Backend to get an access token
        req = self.client.post(
            "/users/get-token",
            json={"pat": api_key},
        )
        json = req.json()
        self.access_token = json["access_token"]
        self.refresh_token = json["refresh_token"]

        if not self.access_token:
            raise Exception("No user found for this API key")

        self.database: supabase.Client = supabase.create_client(
            url,
            key,
            options=ClientOptions(
                schema="datasets",
                auto_refresh_token=True,
                persist_session=False,
            ),
        )
        self.supabase: supabase.Client = supabase.create_client(
            url,
            key,
            options=ClientOptions(
                auto_refresh_token=True,
                persist_session=False,
            ),
        )
        self.database.postgrest.auth(self.access_token)
        self.supabase.postgrest.auth(self.access_token)
        # Set the session on the supabase client
        self.supabase.auth.set_session(self.access_token, self.refresh_token)
        self.database.auth.set_session(self.access_token, self.refresh_token)
        self.user = self.supabase.auth.get_user(self.access_token).user

        log.info(f"Successfully logged in as {self.user.email}.")



# This file is used to import all the resources in the ouro package

from .content import Content, Editor
from .conversations import Conversations
from .datasets import Datasets
from .files import Files
from .posts import Posts

__all__ = ["Content", "Editor", "Conversations", "Datasets", "Posts", "Files"]


def EditorFactory(self, **kwargs) -> Editor:
    return Editor(**kwargs)


def ContentFactory(self, **kwargs) -> Content:
    return Content(**kwargs)


def __init__(self, ouro):
    # Earth
    self.datasets = Datasets(ouro)
    self.files = Files(ouro)

    # Air
    self.posts = Posts(ouro)
    self.conversations = Conversations(ouro)

    self.Editor = self.EditorFactory
    self.Content = self.ContentFactory



import logging
from typing import List, Optional

from ouro._resource import SyncAPIResource
from ouro.models import Conversation

from .content import Content

log: logging.Logger = logging.getLogger(__name__)


__all__ = ["Conversations", "Messages"]


class Messages(SyncAPIResource):
    def create(self, conversation_id: str, **kwargs):
        json = kwargs.get("json")
        text = kwargs.get("text")
        user_id = kwargs.get("user_id")
        message = {
            "json": json,
            "text": text,
            "user_id": user_id,
            **kwargs,
        }

        message = {k: v for k, v in message.items() if v is not None}
        request = self.client.post(
            f"/conversations/{conversation_id}/messages/create",
            json={"message": message},
        )
        request.raise_for_status()
        response = request.json()
        if response["error"]:
            raise Exception(response["error"])
        return response["data"]

    # def retrieve(self, id: str):
    #     request = self.client.get(f"/messages/{id}")
    #     request.raise_for_status()
    #     response = request.json()
    #     if response["error"]:
    #         raise Exception(response["error"])
    #     return response["data"]

    # def update(self, id: str, content: Optional[Content] = None, **kwargs):
    #     message = {**kwargs}
    #     message = {k: v for k, v in message.items() if v is not None}
    #     request = self.client.put(
    #         f"/messages/{id}",
    #         json={"message": message, "content": content.to_dict() if content else None},
    #     )
    #     request.raise_for_status()
    #     response = request.json()
    #     if response["error"]:
    #         raise Exception(response["error"])
    #     return response["data"]

    # def delete(self, id: str):
    #     request = self.client.delete(f"/messages/{id}")
    #     request.raise_for_status()
    #     response = request.json()
    #     if response["error"]:
    #         raise Exception(response["error"])
    #     return response["data"]

    def list(self, conversation_id: str, **kwargs):
        request = self.client.get(
            f"/conversations/{conversation_id}/messages", params=kwargs
        )
        request.raise_for_status()
        response = request.json()
        if response["error"]:
            raise Exception(response["error"])
        return response["data"]


class ConversationMessages:
    def __init__(self, conversation: "Conversation"):
        self.conversation = conversation
        self.ouro = conversation._ouro

    def create(self, **kwargs):
        return Messages(self.ouro).create(self.conversation.id, **kwargs)

    def retrieve(self, message_id: str):
        return Messages(self.ouro).retrieve(message_id)

    def update(self, message_id: str, content: Optional[Content] = None, **kwargs):
        return Messages(self.ouro).update(message_id, content, **kwargs)

    def delete(self, message_id: str):
        return Messages(self.ouro).delete(message_id)

    def list(self, **kwargs):
        return Messages(self.ouro).list(self.conversation.id, **kwargs)


class Conversations(SyncAPIResource):
    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    #     self.messages = Messages(*args, **kwargs)

    def create(
        self,
        **kwargs,
    ) -> Conversation:
        """
        Create a new Conversation
        """

        conversation = {
            **kwargs,
        }
        # Filter out None values
        conversation = {k: v for k, v in conversation.items() if v is not None}

        request = self.client.post(
            "/conversations/create",
            json={
                **conversation,
            },
        )
        request.raise_for_status()
        response = request.json()
        if response["error"]:
            raise Exception(response["error"])

        return Conversation(**response["data"], _ouro=self.ouro)

    def retrieve(self, id: str) -> Conversation:
        """
        Retrieve a Conversation by its id
        """
        request = self.client.get(
            f"/conversations/{id}",
        )
        request.raise_for_status()
        response = request.json()
        if response["error"]:
            raise Exception(response["error"])

        return Conversation(**response["data"], _ouro=self.ouro)

    # def update(
    #     self,
    #     id: str,
    #     name: Optional[str] = None,
    #     description: Optional[str] = None,
    #     visibility: Optional[str] = None,
    #     **kwargs,
    # ) -> Conversation:
    #     """
    #     Update a Conversation by its id
    #     """
    #     conversation = {
    #         "name": name,
    #         "description": description,
    #         "visibility": visibility,
    #         **kwargs,
    #     }
    #     conversation = {k: v for k, v in conversation.items() if v is not None}

    #     request = self.client.put(
    #         f"/conversations/{id}",
    #         json=conversation,
    #     )
    #     request.raise_for_status()
    #     response = request.json()
    #     if response["error"]:
    #         raise Exception(response["error"])
    #     return Conversation(**response["data"])

    # def delete(self, id: str):
    #     """
    #     Delete a Conversation by its id
    #     """
    #     request = self.client.delete(
    #         f"/conversations/{id}",
    #     )
    #     request.raise_for_status()
    #     response = request.json()
    #     if response["error"]:
    #         raise Exception(response["error"])

    #     return response["data"]



def ouro_field(key, value):
    """
    Decorator to add custom fields to the OpenAPI schema of your FastAPI app.
    """

    def decorator(func):
        if not hasattr(func, "ouro_fields"):
            func.ouro_fields = {}
        func.ouro_fields[key] = value
        return func

    return decorator


def get_custom_openapi(app, get_openapi):
    """
    Function to generate a custom OpenAPI schema for your FastAPI app.
    """

    def custom_openapi():
        if app.openapi_schema:
            return app.openapi_schema

        openapi_schema = get_openapi(
            title=app.title,
            version=app.version,
            summary=app.summary,
            description=app.description,
            routes=app.routes,
        )

        for path, path_item in openapi_schema["paths"].items():
            for method, operation in path_item.items():
                endpoint = app.routes[0]
                for route in app.routes:
                    if route.path == path and method.upper() in route.methods:
                        endpoint = route.endpoint
                        break

                if hasattr(endpoint, "ouro_fields"):
                    operation.update(endpoint.ouro_fields)

        app.openapi_schema = openapi_schema
        return app.openapi_schema

    return custom_openapi



from datetime import datetime
from typing import TYPE_CHECKING, List, Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

if TYPE_CHECKING:
    from ouro.resources.conversations import ConversationMessages

    from ouro import Ouro

__all__ = [
    "Asset",
    "PostContent",
    "Post",
    "Conversation",
]


class Asset(BaseModel):
    id: UUID
    user_id: UUID
    org_id: UUID | None
    name: str
    visibility: str
    asset_type: str
    created_at: datetime
    last_updated: datetime
    description: Optional[str]
    metadata: Optional[dict]
    monetization: Optional[str]
    price: Optional[float]
    product_id: Optional[str]
    price_id: Optional[str]
    preview: Optional[dict]
    cost_accounting: Optional[str]
    cost_unit: Optional[str]
    unit_cost: Optional[float]


class PostContent(BaseModel):
    text: str
    data: dict = Field(
        alias="json",
    )


class Post(Asset):
    content: Optional[PostContent] = None
    # preview: Optional[PostContent]
    comments: Optional[int] = Field(default=0)
    views: Optional[int] = Field(default=0)


class ConversationMetadata(BaseModel):
    members: List[UUID]
    summary: Optional[str] = None


class Conversation(Asset):
    id: UUID
    name: str
    description: Optional[str] = None
    summary: Optional[str] = None
    metadata: ConversationMetadata
    _messages: Optional["ConversationMessages"] = None
    _ouro: Optional["Ouro"] = None

    def __init__(self, _ouro=None, **data):
        super().__init__(**data)
        self._ouro = _ouro

    @property
    def messages(self):
        if self._messages is None:
            from ouro.resources.conversations import ConversationMessages

            self._messages = ConversationMessages(self)
        return self._messages



import json
import logging
import mimetypes
import os
import uuid
from typing import List, Optional

from ouro._resource import SyncAPIResource

log: logging.Logger = logging.getLogger(__name__)


__all__ = ["Files"]


class Files(SyncAPIResource):
    def create(
        self,
        file_path: str,
        name: str,
        visibility: str,
        monetization: Optional[str] = None,
        price: Optional[float] = None,
        description: Optional[str] = None,
    ):

        # Refresh supabase client
        self.ouro.supabase.auth.refresh_session()

        log.debug(f"Creating a file")
        # Update file with Supabase
        with open(file_path, "rb") as f:
            # Get file extension and MIME type
            mime_type = mimetypes.guess_type(file_path)[0]
            file_extension = os.path.splitext(file_path)[1]
            id = str(uuid.uuid4())

            bucket = "public-files" if visibility == "public" else "files"
            path_on_storage = f"{self.ouro.user.id}/{id}{file_extension}"

            print(f"Uploading file to {path_on_storage} in the {bucket} bucket")
            request = self.ouro.supabase.storage.from_(bucket).upload(
                file=f,
                path=path_on_storage,
                file_options={"content-type": mime_type},
            )
            request.raise_for_status()
            file = request.json()
            print(file)

        # Not sure why it's cased like this
        file_id = file["Id"]

        # Get file details server-side
        request = self.client.get(
            f"/files/{file_id}/metadata",
        )
        request.raise_for_status()
        response = request.json()

        metadata = response["data"]["metadata"]
        metadata = {
            "name": f"{id}{file_extension}",
            "bucket": bucket,
            "path": path_on_storage,
            "type": mime_type,
            "mimeType": mime_type,
            **metadata,
        }
        preview = response["data"]["preview"]

        body = {
            "id": file_id,
            "name": name,
            "visibility": visibility,
            "monetization": monetization,
            "price": price,
            "description": description,
            "metadata": metadata,
            "preview": preview,
            "asset_type": "file",
        }

        # Filter out None values
        body = {k: v for k, v in body.items() if v is not None}

        request = self.client.post(
            "/datasets/create/from-file",
            json=body,
        )
        request.raise_for_status()
        response = request.json()

        log.info(response)

        if response["error"]:
            raise Exception(json.dumps(response["error"]))

        return response["data"]
